#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_emase
----------------------------------

Tests for `emase` module.
"""

import unittest

from emase import AlignmentPropertyMatrix as APM
from emase import AlignmentMatrixFactory as AMF

class TestEmase(unittest.TestCase):

    def setUp(self):
        pass

    def test_something(self):
        pass

    def tearDown(self):
        pass

if __name__ == '__main__':
    unittest.main()
